import time
from girder import events

from girder.models.file import File
from girder.utility.abstract_assetstore_adapter import AbstractAssetstoreAdapter
from girder_jobs.constants import JobStatus
from girder_jobs.models.job import Job
from girder_worker import GirderWorkerPluginABC

from .models import AssetstoreImport, ImportTrackerCancelError


def wrapShouldImportFile():
    baseShouldImportFile = AbstractAssetstoreAdapter.shouldImportFile

    def shouldImportFileWrapper(self, path, params):
        jobRec = params.get('_job')
        job = None
        if jobRec:
            job = Job().load(jobRec['id'], force=True, includeLog=False)
            if job['status'] == JobStatus.CANCELED:
                raise ImportTrackerCancelError()
            if time.time() - jobRec['lastlog'] > 10:
                Job().updateJob(
                    job,
                    log='%s - Checked %d, skipped %d; checking %s\n' % (
                        time.strftime('%Y-%m-%d %H:%M:%S'),
                        jobRec['count'], jobRec['skip'], path),
                    overwrite=(jobRec['logcount'] > 1000))
                if jobRec['logcount'] > 1000:
                    jobRec['logcount'] = 0
                jobRec['logcount'] += 1
                jobRec['lastlog'] = time.time()
        result = True
        if params.get('excludeExisting'):
            idx1 = ([('assetstoreId', 1), ('path', 1)], {})
            idx2 = ([('assetstoreId', 1), ('s3Key', 1)], {})
            if idx1 not in File()._indices:
                File().ensureIndex(idx1)
            if idx2 not in File()._indices:
                File().ensureIndex(idx2)
            if File().findOne({
                'assetstoreId': self.assetstore['_id'],
                '$or': [{'path': path}, {'s3Key': path}],
                'imported': True
            }):
                result = False
        if result:
            result = baseShouldImportFile(self, path, params)
        if jobRec:
            if not result:
                jobRec['skip'] += 1
            else:
                jobRec['count'] += 1
        return result

    AbstractAssetstoreAdapter.shouldImportFile = shouldImportFileWrapper


def createImportRecord(event: events.Event):
    info = event.info
    path = info['params']['importPath']
    record_data = {
        'destinationId': info['parent']['_id'],
        'destinationType': info['parentType'],
        'importPath': path,
        'leafFoldersAsItems': str(info['params'].get('leafFoldersAsItems', False)).lower(),
        'progress': str(info['progress']).lower(),
        **info['params']
    }

    job = Job().createJob(
        title='Import from %s : %s' % (info['assetstore']['name'], path),
        type='assetstore_import',
        public=False,
        user=info['user'],
        kwargs=info['params'],
    )
    job = Job().updateJob(job, '%s - Starting import from %s : %s\n' % (
        time.strftime('%Y-%m-%d %H:%M:%S'),
        info['assetstore']['name'], path,
    ), status=JobStatus.RUNNING)

    # We mutate the params dict in-place as a side-effect, unfortunately.
    info['params']['_job'] = {
        'id': str(job['_id']),
        'count': 0,
        'skip': 0,
        'lastlog': time.time(),
        'logcount': 0,
    }

    if 'fileIncludeRegex' in info['params']:
        record_data['fileIncludeRegex'] = info['params']['fileIncludeRegex']
    if 'fileExcludeRegex' in info['params']:
        record_data['fileExcludeRegex'] = info['params']['fileExcludeRegex']
    if 'excludeExisting' in info['params']:
        record_data['excludeExisting'] = str(info['params']['excludeExisting']).lower()

    import_record = AssetstoreImport().createAssetstoreImport(info['assetstore'], record_data)
    event.addResponse({'importRecord': import_record})


def finalizeImportRecord(event: events.Event):
    job_info = event.info['params']['_job']
    job = Job().load(job_info['id'], force=True, includeLog=False)

    if 'exception' in event.info:
        exc = event.info['exception']
        if isinstance(exc, ImportTrackerCancelError):
            success = 'canceled'
            Job().updateJob(job, '%s - Canceled\n' % time.strftime('%Y-%m-%d %H:%M:%S'))
        else:
            success = False
            Job().updateJob(job, '%s - Failed with %s' % (
                time.strftime('%Y-%m-%d %H:%M:%S'),
                exc,
            ), status=JobStatus.ERROR)
    else:
        success = True
        Job().updateJob(job, '%s - Finished.  Checked %d, skipped %d\n' % (
            time.strftime('%Y-%m-%d %H:%M:%S'),
            job_info['count'], job_info['skip'],
        ), status=JobStatus.SUCCESS)

    pre_event = event.info['pre_event']
    for response in pre_event.responses:
        if isinstance(response, dict) and 'importRecord' in response:
            AssetstoreImport().markEnded(response['importRecord'], success)
            break


class ImportTrackerWorkerPlugin(GirderWorkerPluginABC):
    def __init__(self, app, *args, **kwargs):
        wrapShouldImportFile()

        events.bind('assetstore_import.before', 'import_tracker', createImportRecord)
        events.bind('assetstore_import.after', 'import_tracker', finalizeImportRecord)
        events.bind('assetstore_import.error', 'import_tracker', finalizeImportRecord)



from bson.objectid import ObjectId

from girder.api import access
from girder.api.describe import Description, autoDescribeRoute
from girder.api.rest import boundHandler
from girder.constants import AccessType, SortDir
from girder.models.assetstore import Assetstore
from girder.models.folder import Folder
from girder.utility import model_importer, path

from . import utils
from .models import AssetstoreImport


def processCursor(cursor, user):
    lookedupAssetstores = {}
    results = list(cursor)

    for row in results:
        if row['assetstoreId'] not in lookedupAssetstores:
            assetstore = list(Assetstore().find({'_id': row['assetstoreId']}))
            lookedupAssetstores[row['assetstoreId']] = assetstore[0][
                'name'] if assetstore else None

        row['_assetstoreName'] = (
            lookedupAssetstores[row['assetstoreId']] or row['assetstoreId'])
        model = model_importer.ModelImporter.model(row['params']['destinationType'])
        doc = model.load(row['params']['destinationId'], user=user)
        if doc:
            row['_destinationPath'] = path.getResourcePath(
                row['params']['destinationType'],
                doc,
                user=user
            )
        else:
            row['_destinationPath'] = 'does not exist'
    return results


def getImports(query=None, user=None, unique=False, limit=None, offset=None, sort=None):
    if not unique:
        cursor = AssetstoreImport().find(
            query or {},
            limit=limit,
            offset=offset,
            sort=sort,
        )
    else:
        cursor = AssetstoreImport().collection.aggregate([
            {'$match': query or {}},
            {'$sort': {k: v for k, v in sort}},
            {'$group': {
                '_id': {
                    'assetstoreId': '$assetstoreId',
                    'params': '$params'
                },
                '_count': {'$sum': 1},
                'started': {'$first': '$started'},
                'ended': {'$first': '$ended'},
                'first_id': {'$first': '$_id'},
            }},
            {'$project': {
                'assetstoreId': '$_id.assetstoreId',
                'params': '$_id.params',
                '_count': '$_count',
                'started': '$started',
                'ended': '$ended',
                '_id': '$first_id',
            }},
            {'$sort': {k: v for k, v in sort}},
            {'$skip': offset or 0},
            {'$limit': limit or 0}
        ])
    return processCursor(cursor, user)


@access.admin
@boundHandler
@autoDescribeRoute(
    Description('List all imports for a given assetstore.')
    .param('id', 'An assetstore ID', paramType='path')
    .param('unique', 'If true, only show unique imports', required=False,
           dataType='boolean', default=False)
    .pagingParams(defaultSort='started', defaultSortDir=SortDir.DESCENDING)
)
def listImports(self, id, unique, limit, offset, sort):
    return getImports(
        {'assetstoreId': ObjectId(id)},
        self.getCurrentUser(), unique, limit, offset, sort)


@access.admin
@boundHandler
@autoDescribeRoute(
    Description('List all past imports for all assetstores.')
    .param('unique', 'If true, only show unique imports', required=False,
           dataType='boolean', default=False)
    .pagingParams(defaultSort='started', defaultSortDir=SortDir.DESCENDING)
)
def listAllImports(self, unique, limit, offset, sort):
    return getImports(
        None,
        self.getCurrentUser(), unique, limit, offset, sort)


@access.admin
@boundHandler
@autoDescribeRoute(
    Description('Get import for record from ID.')
    .modelParam('id', 'ID of an import', model=AssetstoreImport)
)
def getImport(self, assetstoreImport):
    return assetstoreImport


@access.admin
@boundHandler
@autoDescribeRoute(
    Description('Move folder contents to an assetstore.')
    .modelParam('id', 'Source folder ID', model=Folder, level=AccessType.WRITE)
    .modelParam('assetstoreId', 'Destination assetstore ID', model=Assetstore, paramType='formData')
    .param('ignoreImported', 'Ignore files that have been directly imported', dataType='boolean',
           default=True, required=True)
    .param('progress', 'Whether to record progress on the move.', dataType='boolean', default=False,
           required=False)
)
def moveFolder(self, folder, assetstore, ignoreImported, progress):
    user = self.getCurrentUser()
    utils.moveFolder.delay(user, folder, assetstore, ignoreImported, progress)



import os
import re
from setuptools import find_packages, setup


def prerelease_local_scheme(version):
    """Return local scheme version unless building on master in CircleCI.

    This function returns the local scheme version number
    (e.g. 0.0.0.dev<N>+g<HASH>) unless building on CircleCI for a
    pre-release in which case it ignores the hash and produces a
    PEP440 compliant pre-release version number (e.g. 0.0.0.dev<N>).
    """
    from setuptools_scm.version import get_local_node_and_date

    # this regex allows us to publish pypi packages from master, our LTS maintenance branches, and
    # our next major version integration branches
    pattern = r'master|[0-9]+\.x-maintenance|v[0-9]+-integration'

    if re.match(pattern, os.getenv('CIRCLE_BRANCH', '')):
        return ''
    else:
        return get_local_node_and_date(version)


with open('README.md') as readme_file:
    readme = readme_file.read()


setup(
    name='girder-import-tracker',
    use_scm_version={'root': '../..', 'local_scheme': prerelease_local_scheme},
    setup_requires=[
        'setuptools-scm',
        'setuptools-git',
    ],
    description='A Girder plugin for data import tracking',
    long_description=readme,
    long_description_content_type='text/markdown',
    author='Kitware, Inc.',
    author_email='kitware@kitware.com',
    url='https://github.com/girder/girder/tree/master/plugins/import_tracker',
    license='Apache 2.0',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    install_requires=[
        'girder-jobs>=3.0.3',
        'girder>=3.1.23',
    ],
    include_package_data=True,
    keywords='girder-plugin, import_tracker',
    packages=find_packages(exclude=['plugin_tests']),
    zip_safe=False,
    entry_points={
        'girder.plugin': [
            'import_tracker = girder_import_tracker:GirderPlugin'
        ],
        'girder_worker_plugins': [
            'import_tracker = girder_import_tracker.girder_worker_plugin:ImportTrackerWorkerPlugin'
        ],
    }
)
